@@ -1,7 +1,5 @@
 #!/usr/bin/python3
-import gi
-import threading
-from gi.repository import GObject, Gio, GLib
+from gi.repository import Gio, GLib
 import time
 import os
 import datetime
@@ -18,7 +16,7 @@ from pathlib import Path
 import apt
 import apt_pkg
 import aptsources.sourceslist
-import mintcommon.aptdaemon
+import aptkit.simpleclient
 
 import pycurl
 
@@ -130,7 +128,6 @@ class ShowInfoCheck(Check):
         super().__init__(title, "", callback)
         self.result = RESULT_INFO
         self.allow_recheck = False
-        self.allow_continue = True
 
     def do_run(self):
         print("=================")
@@ -177,6 +174,12 @@ class VersionCheck(Check):
                 self.message = _("Your edition of Linux Mint is '%s'. It cannot be upgraded to %s.") % (edition, DESTINATION)
                 return
 
+            # Check architecture
+            if platform.machine() != "x86_64":
+                self.result = RESULT_ERROR
+                self.message = _("Your architecture of Linux Mint is '%s'. It cannot be upgraded.") % platform.machine()
+                return
+
 # Check that the computer is plugged in to AC Power
 class PowerCheck(Check):
 
@@ -196,9 +199,10 @@ class TimeshiftCheck(Check):
 
     def do_run(self):
         if self.get_setting("check-timeshift"):
-            self.result = RESULT_ERROR
-            self.message = _("Perform a Timeshift system snapshot before attempting to upgrade.")
-            self.info.append(_("If the upgrade isn't successful, a system snapshot will allow you to go back in time and revert all the changes."))
+            self.result = RESULT_WARNING
+            self.message = _("If the upgrade isn't successful, a system snapshot will allow you to go back in time and revert all the changes.")
+            self.info.append(_("Press 'Fix' to perform a system snapshot."))
+
             if os.path.exists("/usr/bin/timeshift"):
                 today = datetime.datetime.today().strftime('%Y-%m-%d')
                 if today in subprocess.getoutput("timeshift --list"):
@@ -285,7 +289,7 @@ class APTCacheCheck(Check):
                 return
 
     def install_remove_pkgs(self):
-        apt = mintcommon.aptdaemon.APT(self.window)
+        apt = aptkit.simpleclient.SimpleAPTClient(self.window)
         apt.set_finished_callback(self.on_transaction_finished)
         apt.set_cancelled_callback(self.on_transaction_finished)
         if len(self.pkgs_to_remove) > 0:
@@ -310,6 +314,26 @@ class APTRepoCheck(Check):
         super().__init__(_("Package repositories"), _("Checking package repositories..."), callback)
 
     def do_run(self):
+
+        # Error out if any deb822 repositories are present. We do not yet support them.
+        # python3-apt in bookworm has deb822 support, but the version in bullseye doesn't.
+        deb822_sources = []
+        sources_dir = "/etc/apt/sources.list.d/"
+        if os.path.exists(sources_dir):
+            for f in os.listdir(sources_dir):
+                if f.endswith(".sources"):
+                    deb822_sources.append(os.path.join(sources_dir, f))
+        if len(deb822_sources) > 0:
+            self.result = RESULT_ERROR
+            self.message = _("The following sources are defined in DEB822 format. This format isn't supported yet.")
+            table_list = TableList([""])
+            table_list.show_column_names = False
+            for repo in deb822_sources:
+                table_list.values.append([repo])
+            self.info.append(table_list)
+            self.info.append(_("Backup and then delete these files to continue."))
+            return
+
         apt_pkg.init_config()
         self.sources = aptsources.sourceslist.SourcesList(withMatcher=False)
         self.mint_repos = []
@@ -364,31 +388,34 @@ class APTRepoCheck(Check):
         # Check up-to-date-ness on Mint mirror
         problems = []
         mint_timestamp = self.get_url_last_modified("http://packages.linuxmint.com/db/version")
-        mint_age = None
-        if mint_timestamp != None:
+        if mint_timestamp is None:
+            problems.append(_("%s is unreachable") % "http://packages.linuxmint.com")
+        else:
             mint_date = datetime.datetime.fromtimestamp(mint_timestamp)
             now = datetime.datetime.now()
             mint_age = (now - mint_date).days
             print_output(f"Mint repository last modified on {mint_date}")
-        for repo in self.mint_repos:
-            if "packages.linuxmint.com" in repo.uri:
-                continue
-            timestamp = self.get_url_last_modified("%s/db/version" % repo.uri)
-            if timestamp == None:
-                problems.append(_("%s is unreachable") % repo.uri)
-            elif mint_age > 2:
-                date = datetime.datetime.fromtimestamp(timestamp)
-                offset = (mint_date - date).days
-                if mint_age > 2 and offset > 2:
-                    problems.append(_("%s is not up to date. Switch to a different mirror.") % repo.uri)
+
+            for repo in self.mint_repos:
+                if "packages.linuxmint.com" in repo.uri:
+                    continue
+                timestamp = self.get_url_last_modified("%s/db/version" % repo.uri)
+                if timestamp == None:
+                    # Retry with standard repo layout (for repos which use the mint codenames)
+                    new_dist = repo.dist.replace(ORIGIN_CODENAME, DESTINATION_CODENAME)
+                    url = "%s/dists/%s/Release" % (repo.uri, new_dist)
+                    timestamp = self.get_url_last_modified(url)
+                if timestamp == None:
+                    problems.append(_("%s is unreachable") % repo.uri)
+                elif mint_age > 2:
+                    date = datetime.datetime.fromtimestamp(timestamp)
+                    offset = (mint_date - date).days
+                    if mint_age > 2 and offset > 2:
+                        problems.append(_("%s is not up to date. Switch to a different mirror.") % repo.uri)
 
         # Check the base repos can handle destination codename
         for repo in self.base_repos:
-            if repo.dist == "buster/updates":
-                # special case, the repo syntax changed between LMDE 4 and LMDE 5
-                new_dist = "bullseye-security"
-            else:
-                new_dist = repo.dist.replace(ORIGIN_BASE_CODENAME, DESTINATION_BASE_CODENAME)
+            new_dist = repo.dist.replace(ORIGIN_BASE_CODENAME, DESTINATION_BASE_CODENAME)
             url = "%s/dists/%s/Release" % (repo.uri, new_dist)
             timestamp = self.get_url_last_modified(url)
             if timestamp == None:
@@ -407,6 +434,9 @@ class APTRepoCheck(Check):
 
     def get_url_last_modified(self, url):
         try:
+            if (url.startswith(APT_CACHER_NG_HTTPS)):
+                url = 'https://' + url[len(APT_CACHER_NG_HTTPS):]
+
             c = pycurl.Curl()
             c.setopt(pycurl.URL, url)
             c.setopt(pycurl.CONNECTTIMEOUT, 30)
@@ -465,7 +495,7 @@ class APTForeignCheck(Check):
         super().__init__(_("Foreign packages"), _("Looking for foreign packages..."), callback)
 
     def do_run(self):
-        orphans, self.foreigns = get_foreign_packages(find_orphans=False, find_downgradable_packages=True)
+        orphans, orphan_t64_names, self.foreigns = get_foreign_packages(find_orphans=False, find_downgradable_packages=True)
         if len(self.foreigns) > 0:
             self.result = RESULT_ERROR
             self.message = _("The following packages need to be downgraded back to official versions:")
@@ -489,40 +519,82 @@ class APTForeignCheck(Check):
 # Check APT orphan packages
 class APTOrphanCheck(Check):
 
-    def __init__(self, callback=None):
+    def __init__(self, pre_upgrade_orphans, callback):
+        # return the list found to the pre_upgrade_orphans argument
         super().__init__(_("Orphan packages"), _("Looking for orphan packages..."), callback)
+        self.pre_upgrade_orphans = pre_upgrade_orphans
 
     def do_run(self):
-        if self.get_setting("check-orphans"):
-            self.orphans_to_remove = []
-            orphans, foreigns = get_foreign_packages(find_orphans=True, find_downgradable_packages=False)
-            if len(orphans) > 0:
-                settings = Gio.Settings(schema_id="com.linuxmint.mintupgrade")
-                to_keep = settings.get_strv("orphans-to-keep")
-                for orphan in orphans:
-                    pkg, version = orphan
-                    if pkg.name == "mintupgrade":
-                        continue
-                    if pkg.name not in to_keep:
-                        self.orphans_to_remove.append(pkg.name)
+        self.pre_upgrade_orphans.clear()
 
-                if len(self.orphans_to_remove) > 0:
-                    self.result = RESULT_ERROR
-                    self.message = _("The following packages do not exist in the repositories:")
-                    self.fix = self.remove_orphans
-                    table_list = TableList([""])
-                    table_list.show_column_names = False
-                    for orphan in self.orphans_to_remove:
-                        table_list.values.append([orphan])
-                    self.info.append(table_list)
-                    self.info.append(_("Add the packages you want to keep using the preferences, then press 'Check again'."))
-                    self.info.append(_("Then press 'Fix' to remove the packages listed above."))
-                    return
+        orphans, orphan_t64_names, foreigns = get_foreign_packages(find_orphans=True, find_downgradable_packages=False)
+        if len(orphans) > 0:
+            for orphan in orphans:
+                pkg, version = orphan
+                if pkg.name.startswith("linux-image-"):
+                    continue
+                if pkg.name.startswith("linux-headers-"):
+                    continue
+                self.pre_upgrade_orphans.append(pkg.name)
 
-    def remove_orphans(self):
-        if len(self.orphans_to_remove) > 0:
-            run_command('%s remove --purge %s %s' % (APT_GET, APT_QUIET, " ".join(self.orphans_to_remove)))
+        self.info = []
 
+        if len(self.pre_upgrade_orphans) > 0:
+            self.result = RESULT_INFO
+            self.info = []
+            self.allow_recheck = True
+            self.message = _("The following packages do not exist in the repositories:")
+            table_list = TableList([""])
+            table_list.show_column_names = False
+            for orphan in self.pre_upgrade_orphans:
+                table_list.values.append([orphan])
+            self.info.append(table_list)
+            self.info.append(_("In some rare cases orphan packages can interfere with the upgrade."))
+            self.info.append(_("If you decide to uninstall some of these packages press 'Check again' after their removal."))
+            self.info.append(_("Press 'OK' to continue with the upgrade."))
+            return
+
+        self.result = RESULT_SUCCESS
+
+# Remove newly orphaned pkgs (post upgrade)
+class APTRemoveOrphansCheck(Check):
+
+    def __init__(self, pre_upgrade_orphans, callback):
+        # orphans is the list of orphan pkgs pre-upgrade
+        super().__init__(_("Orphan packages"), _("Removing newly orphaned packages..."), callback)
+        self.pre_upgrade_orphans = pre_upgrade_orphans
+
+    def do_run(self):
+        orphan_pkgs = []
+        orphans, orphan_t64_names, foreigns = get_foreign_packages(find_orphans=True, find_downgradable_packages=False)
+        if len(orphans) > 0:
+            for orphan in orphans:
+                pkg, version = orphan
+                if pkg.name.startswith("linux-image-"):
+                    continue
+                if pkg.name.startswith("linux-headers-"):
+                    continue
+                if pkg.name in NEW_ORPHANS_TO_KEEP:
+                    continue
+                if pkg.name in self.pre_upgrade_orphans:
+                    print(f"Orphan {pkg.name} was already an orphan pre-upgrade, skipping it.")
+                    continue
+                print(f"New orphan detected: {pkg.name}")
+                orphan_pkgs.append(pkg.name)
+
+        if len(orphan_pkgs) > 0:
+            # Remove newly orphaned packages
+            print_output("Removing newly orphaned packages")
+            for removal in orphan_pkgs:
+                # The return code indicates a failure if some packages were not found, so ignore it.
+                run_command('%s purge --yes %s' % (APT_GET, removal))
+
+        if len(orphan_t64_names) > 0:
+            # Upgrade t64 orphans to t64 versions
+            print_output("Migrating t64 orphans")
+            for name in orphan_t64_names:
+                # The return code indicates a failure if some packages were not found, so ignore it.
+                run_command('%s install --yes %s' % (APT_GET, name))
 
 # Switch to the target repositories
 class UpdateReposCheck(Check):
@@ -549,12 +621,15 @@ class UpdateReposCheck(Check):
                 print_output("Switching %s to %s" % (source.uri, source.dist))
             elif ORIGIN_BASE_CODENAME in source.dist:
                 # Base repo
-                if source.dist == "buster/updates":
-                    # special case, the repo syntax changed between LMDE 4 and LMDE 5
-                    source.dist = "bullseye-security"
-                else:
-                    source.dist = source.dist.replace(ORIGIN_BASE_CODENAME, DESTINATION_BASE_CODENAME)
-                print_output("Switching %s to %s" % (source.uri, source.dist))
+                source.dist = source.dist.replace(ORIGIN_BASE_CODENAME, DESTINATION_BASE_CODENAME)
+                print_output(f"Switching {source.uri} to {source.dist}")
+                # In Debian Bookworm (LMDE 6), a new component called "non-free-firmware" is introduced.
+                if "main" in source.comps and \
+                   "contrib" in source.comps and \
+                   "non-free" in source.comps and \
+                   "non-free-firmware" not in source.comps:
+                   source.comps.append("non-free-firmware")
+                   print_output(f"Adding non-free-firmware component to {source.uri}")
             if DESTINATION_BASE_CODENAME in source.dist and "partner" in source.comps:
                 print_output("Disabling partner repo (discontinued).")
                 source.set_enabled(False)
@@ -661,6 +736,8 @@ class SimulateUpgradeCheck(Check):
         download_size = fetcher.fetch_needed
         # additional space needed when all finished
         additional_space_needed = cache._depcache.usr_size
+        if additional_space_needed < 0:
+            additional_space_needed = 0.0
 
         # gather mount info so we calculate free space correctly.
         mounted = []
@@ -878,14 +955,26 @@ class PostUpgradeCheck(Check):
 
         # Install new packages
         print_output("Installing new packages")
-        if len(PACKAGES_ADDITIONS) > 0:
-            if not run_command('%s install --yes --no-install-recommends %s' % (APT_GET, " ".join(PACKAGES_ADDITIONS))):
+        additions = PACKAGES_ADDITIONS
+        if os.getenv("XDG_CURRENT_DESKTOP") != None:
+            desktop = os.getenv("XDG_CURRENT_DESKTOP").upper().replace("X-", "")
+            if desktop == "CINNAMON":
+                print_output("Installing new packages for Cinnamon as well")
+                additions += PACKAGES_ADDITIONS_CINNAMON
+            elif desktop == "MATE":
+                print_output("Installing new packages for MATE as well")
+                additions += PACKAGES_ADDITIONS_MATE
+            elif desktop == "XFCE":
+                print_output("Installing new packages for Xfce as well")
+                additions += PACKAGES_ADDITIONS_XFCE
+        if len(additions) > 0:
+            if not run_command('%s install --yes --no-install-recommends %s' % (APT_GET, " ".join(additions))):
                 self.result = RESULT_ERROR
                 self.message = _("The following packages could not be installed:")
                 table_list = TableList([""])
                 table_list.show_column_names = False
                 cache = apt.Cache()
-                for name in PACKAGES_ADDITIONS:
+                for name in additions:
                     if name in cache:
                         pkg = cache[name]
                         if not pkg.is_installed:
@@ -897,10 +986,6 @@ class PostUpgradeCheck(Check):
 
         # Install kernel meta
         print_output("Installing kernel meta")
-        if platform.machine() == "x86_64":
-            KERNEL_META = KERNEL_META_64
-        else:
-            KERNEL_META = KERNEL_META_32
         if len(KERNEL_META) > 0:
             if not run_command('%s install --yes --no-install-recommends %s' % (APT_GET, " ".join(KERNEL_META))):
                 self.result = RESULT_ERROR
@@ -932,6 +1017,14 @@ class PostUpgradeCheck(Check):
         print_output("Running apt-get clean")
         run_command("%s clean" % APT_GET)
 
+        # Remove /etc/lsb-release (LMDE only, only /etc/os-release is present in Debian 12+)
+        if IS_LMDE and os.path.exists("/etc/lsb-release"):
+            run_command("rm -f /etc/lsb-release")
+
+        # Update initramfs
+        print_output("Running update-initramfs -u")
+        run_command("update-initramfs -u")
+
         # Adjust Grub title
         if os.path.exists("/usr/share/ubuntu-system-adjustments/systemd/adjust-grub-title"):
             os.system("/usr/share/ubuntu-system-adjustments/systemd/adjust-grub-title")
